// Copyright 2013 Michael E. Stillman.

#ifndef _aring_qq_hpp_
#define _aring_qq_hpp_

#include "aring-qq-flint.hpp"
#include "aring-qq-gmp.hpp"

namespace M2 {
class ARingQQFlint;
class ARingQQGMP;

typedef ARingQQGMP ARingQQ;
};

#endif

// Local Variables:
// compile-command: "make -C $M2BUILDDIR/Macaulay2/e "
// indent-tabs-mode: nil
// End:
