
# Here are a couple of options to further analyse the result
#
# The number of finite solutions:
#
#   nfinite(result)
#
# The number of real solutions:
#
#   nreal(result)
#
# All real solutions
#
#   real(result)
