// Copyright 2015 Anton Leykin and Mike Stillman

// Anton Leykin's code in this file is in the public domain.

#ifndef _slp_hpp_
#define _slp_hpp_

#include "mutablemat-defs.hpp"
#include "SLP-defs.hpp"
#include "SLP-imp.hpp"
#include "mutablemat-imp.hpp"

#include "buffer.hpp"
#include "matrix.hpp"
#include "aring-glue.hpp"

#endif

// Local Variables:
// compile-command: "make -C $M2BUILDDIR/Macaulay2/e "
// indent-tabs-mode: nil
// End:
