// Copyright 2005-2012  Michael E. Stillman

#ifndef _mutable_mat_hpp_
#define _mutable_mat_hpp_

#include "mutablemat-defs.hpp"
#include "SLP-defs.hpp"
#include "mutablemat-imp.hpp"
#include "SLP-imp.hpp"
#endif

// Local Variables:
// compile-command: "make -C $M2BUILDDIR/Macaulay2/e "
// indent-tabs-mode: nil
// End:
