 

/**
 * \defgroup rings rings
 * 
 */


/**
 * \defgroup ringinfo ring info
 *  \ingroup rings
 */

/**
 * \defgroup coeffrings coefficient rings
 *  \ingroup rings 
 */

/**
 * \defgroup polynomialrings polynomial rings 
 *  \ingroup rings
 */

/**
 * \defgroup matrices matrices
 * 
 */


/**
 * \defgroup comp Computations
 * 
 */

/**
 * \defgroup gb Groebner Bases
 *  \ingroup comp
 * 
 */

/**
 * \defgroup res Free Resolutions
 *  \ingroup comp
 * 
 */

/**
 * \defgroup hilb Hilbert Functions
 *  \ingroup comp
 * 
 */
